#include <linux/kernel.h>
#include <linux/init.h>
#include <linux/pci.h>
#include <linux/interrupt.h>

#include <asm/mach/pci.h>
#include <asm/arch/ns9750_pci.h>

#define PCI_NAME "ns9750_pci"

#undef DEBUG
//#define DEBUG
#if defined(DEBUG)
#define DEBUG_FUNC() printk(PCI_NAME ": %s()\n", __FUNCTION__)
#define DEBUG_INFO(args...) printk(PCI_NAME ": " args)
#else
#define DEBUG_FUNC()
#define DEBUG_INFO(args...)
#endif

#define PRINT_WARN(args...) printk(KERN_WARNING PCI_NAME ": " args)

#define NS9750_PCI_CA_BASE (0xfa100000)
#define NS9750_PCI_CD_BASE (0xfa200000)

#define CONFIG_CMD(bus, devfn, where) \
	(0x80000000 | (bus->number << 16) | (devfn << 8) | (where & ~3))

static int
ns9750_read_config(struct pci_bus *bus, unsigned int devfn, int where,
                      int size, u32 *value)
{
	if(devfn & 0xf0){
		*value = 0;
		return PCIBIOS_SUCCESSFUL;
	}
	outl(CONFIG_CMD(bus, devfn, where), NS9750_PCI_CA_BASE);
	switch(size){
	case 1:
		*value = inb(NS9750_PCI_CD_BASE + (where & 3));
		break;
	case 2:
		*value = inw(NS9750_PCI_CD_BASE + (where & 2));
		break;
	case 4:
		*value = inl(NS9750_PCI_CD_BASE);
		break;
	default:
		*value = 0;
		break;
	}
	DEBUG_INFO("[RD:%d] *0x%08x = 0x%08x\n",
		   size, CONFIG_CMD(bus, devfn, where) + (where & 3), *value);
	return PCIBIOS_SUCCESSFUL;
}

static int
ns9750_write_config(struct pci_bus *bus, unsigned int devfn, int where,
                       int size, u32 value)
{
	if(devfn & 0xf0){
		return PCIBIOS_SUCCESSFUL;
	}

	outl(CONFIG_CMD(bus, devfn, where), NS9750_PCI_CA_BASE);
	switch(size){
	case 1:
		outb(value, NS9750_PCI_CD_BASE + (where & 3));
		break;
	case 2:
		outw(value, NS9750_PCI_CD_BASE + (where & 2));
		break;
	case 4:
		outl(value, NS9750_PCI_CD_BASE);
		break;
	default:
		break;
	}
	DEBUG_INFO("[WR:%d] *0x%08x = 0x%08x\n",
		   size, CONFIG_CMD(bus, devfn, where) + (where & 3), value);
	return PCIBIOS_SUCCESSFUL;
}

static int irq_table[5] __initdata = {
	0, IRQ_PCI_EXT_0, IRQ_PCI_EXT_1, IRQ_PCI_EXT_2, IRQ_PCI_EXT_3
};

static int __init
ns9750_map_irq(struct pci_dev *dev, u8 slot, u8 pin)
{
	DEBUG_FUNC();
	return irq_table[(slot + (pin - 1)) & 3];
}

static struct pci_ops ns9750_ops = {
	.read	= ns9750_read_config,
	.write	= ns9750_write_config,
};

static struct pci_bus * __init
ns9750_scan_bus(int nr, struct pci_sys_data *sys)
{
	DEBUG_FUNC();
	return pci_scan_bus(0, &ns9750_ops, sys);
}

static struct resource io_resource = {
	.name	= "PCI io",
	.start	= (0xfa000000), //NS9750_PCI_IO_BASE,
	.end	= (0xfa000000) + (0x00100000) - 1, //NS9750_PCI_IO_BASE + NS9750_PCI_IO_SIZE - 1,
	.flags	= IORESOURCE_IO,
};

static struct resource mem_resource = {
	.name	= "PCI mem",
	.start	= (0x80000000),//NS9750_PCIMEM_BASE,
	.end	= (0x80000000) + (0x10000000) - 1, //NS9750_PCIMEM_BASE + NS9750_PCIMEM_SIZE - 1,
	.flags	= IORESOURCE_MEM,
};

static int __init
ns9750_setup(int nr, struct pci_sys_data *sys)
{
	DEBUG_FUNC();

	if(!request_region(NS9750_PCI_CA_BASE,
			    0x00100000, "pci config address")){
		PRINT_WARN("Unable to request region pci config address\n");
		return 0;
	}
	if (!request_region(NS9750_PCI_CD_BASE,
			    0x00100000, "pci config data")){
		PRINT_WARN("Unable to request region pci config data\n");
		release_region(NS9750_PCI_CA_BASE, 0x00100000);
		return 0;
	}

	sys->resource[0] = &io_resource;
	sys->resource[1] = &mem_resource;

	return 1;
}

static void __init
ns9750_preinit(void)
{
	DEBUG_FUNC();

	writel( NS_PCI_EN_BAR0, NS_PCI_PMISC );

	/* Set up bridge configuration in CMD register */
	writel( 0x80000004, NS_PCI_CONFIG_ADDR );
	writel( 0x00000046, NS_PCI_CONFIG_DATA );

	/* config BAR0: 0x00000000, 256 Mbytes */
	writel( 0x80000010, NS_PCI_CONFIG_ADDR );
	writel( 0x00000000, NS_PCI_CONFIG_DATA );
  
	/* Set up memory space translation */
	writel( 0x43424140, NS_PCI_PALTMEM0 );
	writel( 0x47464544, NS_PCI_PALTMEM1 );
  
	writel( 0xfa000000 >> 20, NS_PCI_PALTIO );
	writel( 0x0, NS_PCI_PMALT0 );
	writel( 0x3, NS_PCI_PALTCTL );

	/* config latency and cache size registers */
	writel( 0x8000000C, NS_PCI_CONFIG_ADDR );
	writel( 0x0000FF00, NS_PCI_CONFIG_DATA );
}

static void __init
ns9750_postinit(void)
{
	DEBUG_FUNC();

	writel( 0x2, NS_PCI_PARBCFG );
}

static struct hw_pci ns9750_pci __initdata = {
	.nr_controllers	= 1,
	.swizzle	= pci_std_swizzle,
	.setup		= ns9750_setup,
	.map_irq	= ns9750_map_irq,
	.scan		= ns9750_scan_bus,
	.preinit	= ns9750_preinit,
	.postinit	= ns9750_postinit,
};

static int __init ns9750_pci_init(void)
{
	DEBUG_FUNC();

	pci_common_init(&ns9750_pci);

	return 0;
}

subsys_initcall(ns9750_pci_init);
