/***********************************************************************
 *
 * Copyright (C) 2004-2005 by FS Forth-Systeme GmbH.
 * All rights reserved.
 *
 * $Id: i2c_ns.c,v 1.5 2005/09/21 16:00:55 jdietsch Exp $
 * @Author: Markus Pietrek
 * @Descr: I2C driver for the NS9750. No bus driver is needed as these are
 *         cpu interal registers.
 *         Slave mode is not supported.
 * @References: [1] NS9750 Hardware Reference, March 2004, Chap.11
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of
 * the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston,
 * MA 02111-1307 USA
 *
 ***********************************************************************/
/***********************************************************************
 *  @History:
 *     2005/03/01 : NS9360 support added
 *     2004/08/17 : removed addr_
 ***********************************************************************/
#include <linux/config.h>	/* system configuration */

#include <asm/arch/ns9750_i2c.h> /* NS_I2C_MODULE_BASE */
#include <asm/arch/ns9750_bbus.h> /* ns_gpio_lock */

#include <asm/arch/irqs.h>	/* IRQ_BBUS_I2C */
#include <linux/module.h>	/* module_init */
#include <linux/init.h>		/* __init */
#include <linux/ioport.h>	/* request_mem_region */
#include <linux/interrupt.h>	/* request_irq */
#include <linux/wait.h>		/* DECLARE_WAIT_QUEUE_HEAD */
#include <linux/i2c.h>		/* i2c_algorithm */
#include <linux/fs.h>
#include <linux/miscdevice.h>
#include <linux/rtc.h>
#include <asm/io.h>		/* writel */
#include <asm/uaccess.h>
#include <asm/rtc.h>

#define DEBUG_INIT		0x0001
#define DEBUG_MINOR		0x0002
#define DEBUG_RX		0x0004
#define DEBUG_TX		0x0008
#define DEBUG_INT		0x0010
#define DEBUG_ERROR		0x4000
#define DEBUG_ERROR_CRIT	0x8000

static int debug = 0xc000;

#define DEBUG_ARGS( FLG, a0, args... )                              \
({                                                                  \
	if((debug & (FLG)) == (FLG))                                \
		printk(KERN_INFO "%s: " a0, __FUNCTION__, ##args ); \
})
#define DEBUG_FN( FLG )                                                     \
({                                                                          \
	if((debug & (FLG)) == (FLG))                                        \
		printk(KERN_INFO "\r%s:line %d\n", __FUNCTION__, __LINE__); \
})
#define ASSERT( expr, func )                                         \
({                                                                   \
	if(!(expr)) {                                                \
		printk(KERN_ERR "Assertion failed! %s:line %d %s\n", \
		       __FUNCTION__, __LINE__, #expr);               \
		func;                                                \
	}                                                            \
})

#define DRIVER_NAME		"i2c_ns"

/* due to worse setup timing reduced to 50KHz */
#define NS_I2C_CFG_INIT_VAL \
	(((((CPU_CLK_FREQ/4)/100000)/2*2) & NS_I2C_CFG_CLREF_MA) | \
	 ((1 << NS_I2C_CFG_SFW_SH) & NS_I2C_CFG_SFW_MA ))

typedef enum
{
	I2C_NS_INT_AWAITING,
	I2C_NS_INT_OK,
	I2C_NS_INT_RETRY,
	I2C_NS_INT_ERROR
} i2c_ns_int_status_t;

static int i2c_ns_master_xfer( struct i2c_adapter* i2c_adap,
			       struct i2c_msg msgs[], int num );
static u32 i2c_ns_master_func( struct i2c_adapter* adap );
static irqreturn_t i2c_ns_int( int irq, void* dev_id,struct pt_regs* regs );
static int i2c_ns_read( struct i2c_adapter* i2c_adap, int count );
static int i2c_ns_write( struct i2c_adapter* i2c_adap,const char* buf,int count);
static int i2c_ns_send_cmd( struct i2c_adapter* i2c_adap, unsigned int cmd );

static int __init i2c_ns_init( void );
static void i2c_ns_exit( void );

DECLARE_WAIT_QUEUE_HEAD( wait_q );

/* no multiple instances possible, don't use struct for local variables */
static spinlock_t i2c_ns_lock;	/* locking for i2c_ns_buf etc. */
static char* i2c_ns_buf = NULL;	/* where to store the incoming character */
static i2c_ns_int_status_t i2c_ns_int_status;

static struct i2c_algorithm i2c_ns_algo = {
	.name		= "9750 I2C algorithm",
	/* register for id @ http://secure.netroedge.com/~lm78/docs.html */
	.id		= I2C_ALGO_NS9750,
	.master_xfer	= i2c_ns_master_xfer,
	.functionality	= i2c_ns_master_func
};

#include "../chips/at24cxx.h"
#include "../chips/s3531a.h"

static struct i2c_client *rtc_client;

static inline int 
rtc_command(int cmd, void *data)
{
	int ret = -EIO;

	if (rtc_client)
		ret = rtc_client->driver->command(rtc_client, cmd, data);

	return ret;
}

static inline int 
rtc_sync(void)
{
	return rtc_command(RTC_SYNC, NULL);
}

static inline int 
get_rtc_time(struct rtc_time *rtctm)
{
	struct s3531a_time rawtm;
	int ret;

	memset(rtctm, 0, sizeof(struct rtc_time));

	ret = rtc_command(RTC_GETDATETIME, &rawtm);
	if (!ret) {
		rtctm->tm_year = rawtm.year + ((rawtm.year <= 69) ? 100 : 0);
		rtctm->tm_mon  = rawtm.mon - 1;
		rtctm->tm_mday = rawtm.mday;
		rtctm->tm_wday = rawtm.wday;
		rtctm->tm_hour = rawtm.hour;
		rtctm->tm_min  = rawtm.min;
		rtctm->tm_sec  = rawtm.sec;
	}

	return ret;
}

static inline int
set_rtc_time(struct rtc_time *rtctm)
{
	struct s3531a_time rawtm;

	rawtm.year = rtctm->tm_year - ((rtctm->tm_year < 100) ? 0 : 100);
	rawtm.mon  = rtctm->tm_mon + 1;
	rawtm.mday = rtctm->tm_mday;
	rawtm.wday = rtctm->tm_wday;
	rawtm.hour = rtctm->tm_hour;
	rawtm.min  = rtctm->tm_min;
	rawtm.sec  = rtctm->tm_sec;

	return rtc_command(RTC_SETDATETIME, &rawtm);
}

static int 
rtc_ioctl(struct inode *inode, struct file *file,
	  unsigned int cmd, unsigned long arg)
{
	struct rtc_time rtctm;
	int ret;

	switch (cmd) {
	case RTC_ALM_READ:
		break;

	case RTC_ALM_SET:
		break;

	case RTC_RD_TIME:
		ret = get_rtc_time(&rtctm);
		if (ret)
			return ret;

		ret = (copy_to_user((void *)arg, &rtctm, sizeof(rtctm)) ? 
		       -EFAULT : 0);
		return ret;

	case RTC_SET_TIME:
		if (copy_from_user(&rtctm,
				   (struct rtc_time*)arg,
				   sizeof(rtctm)))
			return -EFAULT;

		return set_rtc_time(&rtctm);
	}

	return -EINVAL;
}

static struct file_operations rtc_fops = {
	ioctl: rtc_ioctl,
};

static struct miscdevice rtc_dev = {
	.minor	= RTC_MINOR,
	.name	= "rtc",
	.fops	= &rtc_fops,
};

static struct i2c_client *eeprom_client = NULL;

int
eeprom_ioctl(struct inode *inode, struct file *file,
	     unsigned int cmd, unsigned long arg)
{
	int ret = -EINVAL;
	if (eeprom_client)
		ret = eeprom_client->driver->command(eeprom_client,
						     cmd,
						     (void*)arg);

	return ret;
}

static struct file_operations eeprom_fops = {
	ioctl: eeprom_ioctl,
};

static int
ns9750_client_reg(struct i2c_client *client)
{
	struct rtc_time now;
	struct timespec time;

	DEBUG_FN(DEBUG_MINOR);

	switch (client->driver->id) {
	case I2C_DRIVERID_S3531A:
		if (client->addr == I2C_S3531A_ADDRESS && !rtc_client) {
		  rtc_client = client;
		  if (rtc_sync()) {
		  	printk(KERN_WARNING ": RTC stopped.\n");
			rtc_client = NULL;
			return -1;
		  }
		  if (get_rtc_time(&now)) {
		  	printk(KERN_WARNING ": RTC read failed.\n");
			rtc_client = NULL;
			return -1;
		  }
		  time.tv_nsec = 0;
		  rtc_tm_to_time(&now, &time.tv_sec);
		  do_settimeofday(&time);
		}
		break;
	case I2C_DRIVERID_AT24Cxx:
		if ((client->addr == I2C_EEPROM_ADDRESS) && !eeprom_client)
			eeprom_client = client;
		break;
	}

	return 0;
}

static int
ns9750_client_unreg(struct i2c_client *client)
{
	DEBUG_FN(DEBUG_MINOR);
	if ((client == rtc_client) && (client->addr == I2C_S3531A_ADDRESS))
		rtc_client = NULL;

	if ((client == eeprom_client) && (client->addr == I2C_EEPROM_ADDRESS))
		eeprom_client = NULL;

	return 0;
}

static struct i2c_adapter i2c_ns_adapter = {
	.owner			= THIS_MODULE,
	.name			= DRIVER_NAME,
	.id			= I2C_HW_NS9750,
	.algo			= &i2c_ns_algo,
	.client_register	= ns9750_client_reg,
	.client_unregister	= ns9750_client_unreg,
	.timeout		= HZ * 5,
	.retries		= 1
};

/***********************************************************************
 * @Function: i2c_ns_master_xfer
 * @Return: 0 on failure otherwise number messages sends (== num)
 * @Descr: performs action described in i2c_msg
 *         accepts I2C_M_NOSTART for continous transfer, I2C_M_RD and
 *         !I2C_M_RD (write). No buffer validity check is performed.
 ***********************************************************************/

static int i2c_ns_master_xfer( struct i2c_adapter* i2c_adap,
			       struct i2c_msg msgs[], int num )
{
	int i = 0;
	int ret = 0;
	int len;
	char* buf;
	unsigned long flags;
	int retry = 100;		/* allow only some retries */

	DEBUG_FN( DEBUG_MINOR );

	i2c_ns_int_status = I2C_NS_INT_OK;
	while( i < num ) {
		if( i2c_ns_int_status == I2C_NS_INT_RETRY ) {
			/* resend complete message, stop previous access first */
			/* i is already set to 0 due to while condition */
			ret=i2c_ns_send_cmd(i2c_adap,NS_I2C_DATA_TX_CMD_M_STOP);
			if( ret || !(retry--) ) {
				DEBUG_ARGS( DEBUG_ERROR,
					    "max. number of retries failed\n" );

				break;
			}
		}

		len = msgs[ i ].len;
		buf = msgs[ i ].buf;

		spin_lock_irqsave( &i2c_ns_lock, flags );
		/* being modified by int handler */
		i2c_ns_buf = buf;
		spin_unlock_irqrestore( &i2c_ns_lock, flags );

		if( !(msgs[ i ].flags & I2C_M_NOSTART ) ) {
			unsigned int cmd;

			/* set device address */
			writel( ((msgs[i].addr << NS_I2C_MASTER_MDA_SH) &
				 NS_I2C_MASTER_MDA_MA) |
				((msgs[i].flags&I2C_M_TEN)?NS_I2C_MASTER_MAM:0),
				NS_I2C_MASTER );

			if( msgs[ i ].flags & I2C_M_RD )
				cmd = NS_I2C_DATA_TX_CMD_M_READ;
			else {
				/* we need to send first byte with the command */
				cmd = NS_I2C_DATA_TX_CMD_M_WRITE |
					NS_I2C_DATA_TX_VAL |
					*buf;
				len--;
				buf++;
			}

			/* send start, device address and read/write command */
			ret = i2c_ns_send_cmd( i2c_adap, cmd );
			if( ret ) {
				if( i2c_ns_int_status == I2C_NS_INT_RETRY ) {
					/* try the message again */
					i = 0;
					continue;
				}

				DEBUG_ARGS( DEBUG_ERROR,
					    "Setting address failed for 0x%x "
					    "with errno %i\n",
					    msgs[ i ].addr, ret  );
				break;
			}
		}
		if( len > 0 ) {
			/* either read mode or send more than one byte */
			if( msgs[ i ].flags & I2C_M_RD )
				ret = i2c_ns_read( i2c_adap, len );
			else
				ret = i2c_ns_write( i2c_adap, buf, len );

			if( ret ) {
				if( i2c_ns_int_status == I2C_NS_INT_RETRY ) {
					/* try the message again */
					i = 0;
					continue;
				}

				DEBUG_ARGS( DEBUG_ERROR,
					    "Operation 0x%x failed for 0x%x "
					    "with errno %i\n",
					    msgs[ i ].flags,msgs[ i ].addr,ret);
				break;
			}
		}

		i++;
	}

	/* message block transmitted, send stop */
	if( i2c_ns_send_cmd( i2c_adap, NS_I2C_DATA_TX_CMD_M_STOP ) ) {
		/* something locked up, maybe devices thinks lost clock. */
		if( i2c_ns_send_cmd( i2c_adap, NS_I2C_DATA_TX_CMD_M_NOP ) ||
		    i2c_ns_send_cmd( i2c_adap, NS_I2C_DATA_TX_CMD_M_STOP ) ) {
			unsigned long flags;

			/* maybe netsilicon's I2C state machine got stuck,
			   reset it. */
			spin_lock_irqsave( &ns_gpio_lock, flags );
			rmwl( NS_BBUS_MASTER_RESET, | NS_BBUS_MASTER_RESET_I2C );
			rmwl( NS_BBUS_MASTER_RESET, & ~NS_BBUS_MASTER_RESET_I2C);
			spin_unlock_irqrestore( &ns_gpio_lock, flags );
			writel( NS_I2C_CFG_INIT_VAL, NS_I2C_CFG );
		}
	}

	i2c_ns_buf = NULL;	/* maybe we get spurios interrupts */

	return ret ? 0 : i;
}

static u32 i2c_ns_master_func( struct i2c_adapter* adap )
{
	DEBUG_FN( DEBUG_MINOR );

	return (I2C_FUNC_I2C | I2C_FUNC_10BIT_ADDR |
		I2C_FUNC_SMBUS_QUICK | I2C_FUNC_SMBUS_BYTE |
		I2C_FUNC_SMBUS_BYTE_DATA | I2C_FUNC_SMBUS_WORD_DATA);
}

/***********************************************************************
 * @Function: i2c_ns_int
 * @Return: IRQ_HANDLED and i2c_ns_int_status = ERROR or OK
 * @Descr: when data read, stores it in buffer pointed to by i2c_ns_buf. Buffer
 *         is not incremented.
 ***********************************************************************/

static irqreturn_t i2c_ns_int( int irq, void* dev_id,struct pt_regs* regs )
{
	int status;

	DEBUG_FN( DEBUG_INT );

	/* read and acknowledge (by read) */
	status = readl( NS_I2C_DATA_RX );

	switch( status & NS_I2C_DATA_RX_IRQCD_MA ) {
	    case NS_I2C_DATA_RX_IRQCD_M_RX_DATA:
		spin_lock( &i2c_ns_lock );
		if( i2c_ns_buf != NULL )
			/* be paranoid for spurios interrupts? */
			*i2c_ns_buf = status & NS_I2C_DATA_RX_DATA_MA;
		spin_unlock( &i2c_ns_lock );
		/* no break, will be done by M_CMD_ACK */

	    case NS_I2C_DATA_RX_IRQCD_M_CMD_ACK:
		i2c_ns_int_status = I2C_NS_INT_OK;
		break;

	    case NS_I2C_DATA_RX_IRQCD_M_TX_DATA:
		i2c_ns_int_status = I2C_NS_INT_OK;
		break;

	    case NS_I2C_DATA_RX_IRQCD_M_NO_ACK:
		/* maybe device didn't listen closely, try again */
		i2c_ns_int_status = I2C_NS_INT_RETRY;
		break;

	    default:
		DEBUG_ARGS( DEBUG_ERROR,
			    "interrupt failed with 0x%x\n", status );
		i2c_ns_int_status = I2C_NS_INT_ERROR;
	}

	wake_up_interruptible( &wait_q );

	return IRQ_HANDLED;
}

/***********************************************************************
 * @Function: i2c_ns_read
 * @Return: <0 in case of failure otherwise 0
 * @Descr: Must be called after an M_READ has been executed and the
 *         interrupt has happened. Byte has already been stored by i2c_ns_int in
 *         i2c_ns_buf but not incremented. Don't do read command in i2c_ns_int
 *         so that we still use i2c_adap->timeout.
 ***********************************************************************/

static int i2c_ns_read( struct i2c_adapter* i2c_adap, int count )
{
	int ret = 0;
	unsigned long flags;

	DEBUG_FN( DEBUG_RX );

	/* first character has already been stored in buffer by interrupt
	 * handler */
	while( count-- > 1 ) {
		spin_lock_irqsave( &i2c_ns_lock, flags );
		i2c_ns_buf++;
		spin_unlock_irqrestore( &i2c_ns_lock, flags );

		ret = i2c_ns_send_cmd( i2c_adap, NS_I2C_DATA_TX_CMD_M_NOP );
		if( ret )
			return ret;
	}

	return 0;
}

/***********************************************************************
 * @Function: i2c_ns_write
 * @Return: <0 in case of failure otherwise 0
 * @Descr: Must be called after an M_WRITE has been executed and the
 *         first character has been sent already with M_WRITE.
 ***********************************************************************/

static int i2c_ns_write( struct i2c_adapter* i2c_adap, const char* buf,int count)
{
	int ret = 0;

	DEBUG_FN( DEBUG_TX );

	while( count-- ) {
		ret = i2c_ns_send_cmd( i2c_adap,
				       (*buf & NS_I2C_DATA_TX_DATA_MA ) |
				       NS_I2C_DATA_TX_CMD_M_NOP |
				       NS_I2C_DATA_TX_VAL );
		if( ret )
			return ret;

		buf++;
	}

	return ret;
}

/***********************************************************************
 * @Function: i2c_ns_send_cmd
 * @Return: -ETIMEDOUT if no interrupt happened in i2c_adap->timeout or -EIO on
 *          an I2C failure
 * @Descr: sends command and waits signal safe for I2C interrupt acknowleding
 *         command or timeout.
 ***********************************************************************/

static int i2c_ns_send_cmd( struct i2c_adapter* i2c_adap, unsigned int cmd )
{
	int ret = 0;

	DEBUG_FN( DEBUG_INT | DEBUG_MINOR );

	/* now wait. Don't use semaphores as these don't support timeouts */
	i2c_ns_int_status = I2C_NS_INT_AWAITING;
	do {
		writel( cmd, NS_I2C_DATA_TX );
		if( !wait_event_interruptible_timeout(
			    wait_q,
			    i2c_ns_int_status != I2C_NS_INT_AWAITING,
			    i2c_adap->timeout ))
		{
			DEBUG_ARGS( DEBUG_ERROR, "timed out\n" );

			/* not interrupted */
			ret = -ETIMEDOUT;
			break;
		}
	} while( i2c_ns_int_status == I2C_NS_INT_AWAITING );

	if( i2c_ns_int_status != I2C_NS_INT_OK )
		ret = -EIO;

	return ret;
}

static int __init i2c_ns_init( void )
{
	DEBUG_FN( DEBUG_INIT );

	printk( KERN_INFO DRIVER_NAME
		": I2C ns9750 driver $Revision: 1.5 $ initializing\n");

	if( request_mem_region( NS_I2C_MODULE_BASE,
				NS_I2C_CFG + 0x4 - NS_I2C_MODULE_BASE,
				DRIVER_NAME ) == NULL ) {
		printk( DRIVER_NAME ": memory not available\n" );
		goto error;
	}

	if( request_irq( IRQ_I2C, i2c_ns_int, 0, DRIVER_NAME, NULL ) ) {
		printk( DRIVER_NAME ": irq %i not available\n", IRQ_I2C );
		goto error_irq;
	}

#ifdef CONFIG_ARCH_NS9360
	rmwl( NS_BBUS_MASTER_RESET, & 0xFFFFFF7F );
	set_gpio_cfg( 70, NS_GPIO_CFG_FUNC_2 );
	set_gpio_cfg( 71, NS_GPIO_CFG_FUNC_2 );
#endif /* CONFIG_ARCH_NS9360 */

	/* no GPIOs and interrupts to configure, 9750 has fixed pins */
	writel( NS_I2C_CFG_INIT_VAL, NS_I2C_CFG );

	i2c_add_adapter( &i2c_ns_adapter );

	if (misc_register(&rtc_dev))
		printk (KERN_WARNING ": Unable to register %s driver.\n",
			rtc_dev.name);

	if (register_chrdev (EEPROM_MAJOR, "eeprom", &eeprom_fops))
		printk (KERN_WARNING  ": Unable to register %s driver.\n",
			"EEPROM");

	return 0;

error_irq:
	release_mem_region( NS_I2C_MODULE_BASE,
			    NS_I2C_CFG + 0x4 - NS_I2C_MODULE_BASE );
error:
	return -ENODEV;
}

static void i2c_ns_exit( void )
{
	DEBUG_FN( DEBUG_INIT );

	misc_deregister(&rtc_dev);
	unregister_chrdev(EEPROM_MAJOR, "eeprom");

	i2c_del_adapter( &i2c_ns_adapter );

	free_irq( IRQ_I2C, NULL );
	release_mem_region( NS_I2C_MODULE_BASE,
			    NS_I2C_CFG + 0x4 - NS_I2C_MODULE_BASE );
}

subsys_initcall( i2c_ns_init );
module_exit( i2c_ns_exit );

MODULE_AUTHOR( "Markus Pietrek <mpietrek@fsforth.de>" );
MODULE_DESCRIPTION( "I2C-Bus for NS9750" );
MODULE_LICENSE( "GPL" );

module_param(debug, int, 0);
MODULE_PARM_DESC( debug, "debug level" );
